@@ -25,28 +25,25 @@
 #include <glib/gi18n.h>
 #include <cairo/cairo.h>
 
-G_DEFINE_TYPE (TaskItem, task_item, GTK_TYPE_EVENT_BOX);
+struct _TaskItem {
+  GtkEventBox  parent;
+  WnckWindow  *window;
+  WnckScreen  *screen;
+  GdkPixbuf   *pixbuf;
+  GdkRectangle area;
+  GTimeVal     urgent_time;
+  guint        blink_timer;
+  gboolean     mouse_over;
+  GdkMonitor  *monitor;
+  WpApplet    *windowPickerApplet;
+};
 
-#define TASK_ITEM_GET_PRIVATE(obj) (G_TYPE_INSTANCE_GET_PRIVATE ((obj),\
-  TASK_TYPE_ITEM, \
-  TaskItemPrivate))
+G_DEFINE_TYPE (TaskItem, task_item, GTK_TYPE_EVENT_BOX)
 
 #define DEFAULT_TASK_ITEM_HEIGHT 26
 //make the TaskItem two pixles wider to allow for space of the border
 #define DEFAULT_TASK_ITEM_WIDTH 28 + 2
 
-struct _TaskItemPrivate {
-    WnckWindow   *window;
-    WnckScreen   *screen;
-    GdkPixbuf    *pixbuf;
-    GdkRectangle area;
-    GTimeVal     urgent_time;
-    guint        blink_timer;
-    gboolean     mouse_over;
-    GdkMonitor  *monitor;
-    WpApplet    *windowPickerApplet;
-};
-
 enum {
     TASK_ITEM_CLOSED_SIGNAL,
     TASK_ITEM_MONITOR_CHANGED,
@@ -59,23 +56,21 @@ static guint task_item_signals[LAST_SIGNAL] = { 0 };
 /* D&D stuff */
 
 enum {
-    TARGET_WIDGET_DRAGGED, /* if this item is dragged */
-    TARGET_ITEM_DRAGGED
+  TARGET_STRING_DRAGGED,
+  TARGET_TASK_ITEM_DRAGGED /* if this item is dragged */
 };
 
 static const GtkTargetEntry drop_types[] = {
-    { "STRING", 0, 0 },
-    { "text/plain", 0, 0},
-    { "text/uri-list", 0, 0},
-    { "widget", GTK_TARGET_OTHER_WIDGET, TARGET_WIDGET_DRAGGED}, //drag and drop target
-    { "item", GTK_TARGET_SAME_WIDGET, TARGET_ITEM_DRAGGED }
+    { (gchar *) "STRING", 0, TARGET_STRING_DRAGGED },
+    { (gchar *) "text/plain", 0, TARGET_STRING_DRAGGED },
+    { (gchar *) "text/uri-list", 0, TARGET_STRING_DRAGGED },
+    { (gchar *) "task-item", GTK_TARGET_OTHER_WIDGET | GTK_TARGET_SAME_APP, TARGET_TASK_ITEM_DRAGGED } // drag and drop target
 };
 
 static const gint n_drop_types = G_N_ELEMENTS(drop_types);
 
 static const GtkTargetEntry drag_types[] = {
-    { "widget", GTK_TARGET_OTHER_WIDGET, TARGET_WIDGET_DRAGGED}, //drag and drop source
-    { "item", GTK_TARGET_SAME_WIDGET, TARGET_ITEM_DRAGGED }
+    { (gchar *) "task-item", GTK_TARGET_OTHER_WIDGET | GTK_TARGET_SAME_APP, TARGET_TASK_ITEM_DRAGGED } // drag and drop source
 };
 
 static const gint n_drag_types = G_N_ELEMENTS(drag_types);
@@ -88,7 +83,7 @@ static void update_hints (TaskItem *item) {
     WnckWindow *window;
     gint x, y, x1, y1;
     widget = GTK_WIDGET (item);
-    window = item->priv->window;
+    window = item->window;
     /* Skip problems */
     if (!WNCK_IS_WINDOW (window)) return;
     if (!GTK_IS_WIDGET (widget)) return;
@@ -126,12 +121,10 @@ static gboolean on_task_item_button_released (
     WnckWindow *window;
     WnckScreen *screen;
     WnckWorkspace *workspace;
-    TaskItemPrivate *priv;
     g_return_val_if_fail (TASK_IS_ITEM(item), TRUE);
-    priv = item->priv;
-    window = priv->window;
+    window = item->window;
     g_return_val_if_fail (WNCK_IS_WINDOW (window), TRUE);
-    screen = priv->screen;
+    screen = item->screen;
     workspace = wnck_window_get_workspace (window);
     /* If we are in a drag and drop action, then we are not activating
      * the window which received a click
@@ -154,21 +147,26 @@ static gboolean on_task_item_button_released (
     return TRUE;
 }
 
-static void task_item_set_visibility (TaskItem *item) {
+static void
+task_item_set_visibility (TaskItem *item)
+{
     WnckScreen *screen;
     WnckWindow *window;
     WnckWorkspace *workspace;
+    gboolean show_all;
+    gboolean show_window;
+
     g_return_if_fail (TASK_IS_ITEM (item));
-    TaskItemPrivate *priv = item->priv;
-    if (!WNCK_IS_WINDOW (priv->window)) {
+    if (!WNCK_IS_WINDOW (item->window)) {
         gtk_widget_hide (GTK_WIDGET (item));
         return;
     }
-    window = priv->window;
-    screen = priv->screen;
+    window = item->window;
+    screen = item->screen;
     workspace = wnck_screen_get_active_workspace (screen);
-    gboolean show_all = wp_applet_get_show_all_windows (priv->windowPickerApplet);
-    gboolean show_window = FALSE;
+    show_all = wp_applet_get_show_all_windows (item->windowPickerApplet);
+    show_window = FALSE;
+
     if (!wnck_window_is_skip_tasklist (window)) {
         if(workspace != NULL) { //this can happen sometimes
             if (wnck_workspace_is_virtual (workspace)) {
@@ -206,39 +204,57 @@ task_item_get_preferred_height (GtkWidget *widget,
     *minimal_height = *natural_height = requisition.height;
 }
 
-static GdkPixbuf *task_item_sized_pixbuf_for_window (
-    TaskItem   *item,
-    WnckWindow *window,
-    gint size)
+static GdkPixbuf *
+task_item_sized_pixbuf_for_window (TaskItem   *item,
+                                   WnckWindow *window,
+                                   gint        size)
 {
-    GdkPixbuf *pbuf = NULL;
+    GdkPixbuf *pixbuf;
+    gint width, height;
+
+    pixbuf = NULL;
     g_return_val_if_fail (WNCK_IS_WINDOW (window), NULL);
-    if (wnck_window_has_icon_name (window)) {
-        const gchar *icon_name = wnck_window_get_icon_name (window);
-        GtkIconTheme *icon_theme = gtk_icon_theme_get_default ();
-        if (gtk_icon_theme_has_icon (icon_theme, icon_name)) {
-            GdkPixbuf *internal = gtk_icon_theme_load_icon (icon_theme,
-                icon_name,
-                size,
-                GTK_ICON_LOOKUP_FORCE_SIZE,
-                NULL
-            );
-            pbuf = gdk_pixbuf_copy (internal);
+
+    if (wnck_window_has_icon_name (window))
+      {
+        const gchar *icon_name;
+        GtkIconTheme *icon_theme;
+
+        icon_name = wnck_window_get_icon_name (window);
+        icon_theme = gtk_icon_theme_get_default ();
+
+        if (gtk_icon_theme_has_icon (icon_theme, icon_name))
+          {
+            GdkPixbuf *internal;
+
+            internal = gtk_icon_theme_load_icon (icon_theme, icon_name, size,
+                                                 GTK_ICON_LOOKUP_FORCE_SIZE,
+                                                 NULL);
+            pixbuf = gdk_pixbuf_copy (internal);
             g_object_unref (internal);
-        }
-    }
-    if (!pbuf) {
-        pbuf = gdk_pixbuf_copy (wnck_window_get_icon (item->priv->window));
-    }
-    gint width = gdk_pixbuf_get_width (pbuf);
-    gint height = gdk_pixbuf_get_height (pbuf);
-    if (MAX (width, height) != size) {
-        gdouble scale = (gdouble) size / (gdouble) MAX (width, height);
-        GdkPixbuf *tmp = pbuf;
-        pbuf = gdk_pixbuf_scale_simple (tmp, (gint) (width * scale), (gint) (height * scale), GDK_INTERP_HYPER);
-        g_object_unref (tmp);
-    }
-    return pbuf;
+          }
+      }
+
+    if (!pixbuf)
+      {
+        pixbuf = gdk_pixbuf_copy (wnck_window_get_icon (item->window));
+      }
+
+    width = gdk_pixbuf_get_width (pixbuf);
+    height = gdk_pixbuf_get_height (pixbuf);
+
+    if (MAX (width, height) != size)
+      {
+        GdkPixbuf *unscaled_pixbuf;
+        gdouble scale;
+
+        scale = (gdouble) size / (gdouble) MAX (width, height);
+        unscaled_pixbuf = pixbuf;
+        pixbuf = gdk_pixbuf_scale_simple (unscaled_pixbuf, (gint) (width * scale), (gint) (height * scale), GDK_INTERP_HYPER);
+        g_object_unref (unscaled_pixbuf);
+      }
+
+    return pixbuf;
 }
 
 /* Callback to draw the icon, this function is responsible to draw the different states of the icon
@@ -249,21 +265,27 @@ static gboolean task_item_draw (
     cairo_t *cr,
     WpApplet* windowPickerApplet)
 {
-    g_return_val_if_fail (widget != NULL, FALSE);
-    g_return_val_if_fail (TASK_IS_ITEM (widget), FALSE);
-    TaskItem *item = TASK_ITEM (widget);
-    TaskItemPrivate *priv = item->priv;
-    g_return_val_if_fail (WNCK_IS_WINDOW (priv->window), FALSE);
-    cr = gdk_cairo_create (gtk_widget_get_window(widget));
+    TaskItem *item;
     GdkRectangle area;
     GdkPixbuf *pbuf;
-    area = priv->area;
-    pbuf = priv->pixbuf;
-    gint size = MIN (area.height, area.width) - 8;
-    gboolean active = wnck_window_is_active (priv->window);
+    gint size;
+    gboolean active;
+    gboolean icons_greyscale;
+    gboolean attention;
+
+    g_return_val_if_fail (widget != NULL, FALSE);
+    g_return_val_if_fail (TASK_IS_ITEM (widget), FALSE);
+    item = TASK_ITEM (widget);
+    g_return_val_if_fail (WNCK_IS_WINDOW (item->window), FALSE);
+    cr = gdk_cairo_create (gtk_widget_get_window(widget));
+    area = item->area;
+    pbuf = item->pixbuf;
+    size = MIN (area.height, area.width) - 8;
+    active = wnck_window_is_active (item->window);
     /* load the GSettings key for gray icons */
-    gboolean icons_greyscale = wp_applet_get_icons_greyscale (priv->windowPickerApplet);
-    gboolean attention = wnck_window_or_transient_needs_attention (priv->window);
+    icons_greyscale = wp_applet_get_icons_greyscale (item->windowPickerApplet);
+    attention = wnck_window_or_transient_needs_attention (item->window);
+
     if (GDK_IS_PIXBUF (pbuf) &&
         gdk_pixbuf_get_width (pbuf) != size &&
         gdk_pixbuf_get_height (pbuf) != size)
@@ -277,7 +299,7 @@ static gboolean task_item_draw (
         cairo_rectangle (cr, area.x + 1, area.y + 1, area.width - 2, area.height - 2);
         cairo_set_source_rgba (cr, .8, .8, .8, .2);
         cairo_fill_preserve (cr);
-        if(priv->mouse_over) {
+        if(item->mouse_over) {
             cairo_set_source_rgba (cr, .9, .9, 1, 0.45);
             cairo_stroke (cr);
         } else {
@@ -285,7 +307,7 @@ static gboolean task_item_draw (
             cairo_set_source_rgba (cr, .8, .8, .8, .4);
             cairo_stroke (cr);
         }
-    } else if(priv->mouse_over) {
+    } else if(item->mouse_over) {
         int glow_x, glow_y;
         cairo_pattern_t *glow_pattern;
         glow_x = area.width / 2;
@@ -300,9 +322,9 @@ static gboolean task_item_draw (
         cairo_paint (cr);
     }
     if (!pbuf) {
-        pbuf = priv->pixbuf = task_item_sized_pixbuf_for_window (item, priv->window, size);
+        pbuf = item->pixbuf = task_item_sized_pixbuf_for_window (item, item->window, size);
     }
-    if (active || priv->mouse_over || attention || !icons_greyscale) {
+    if (active || item->mouse_over || attention || !icons_greyscale) {
         gdk_cairo_set_source_pixbuf (
             cr,
             pbuf,
@@ -334,15 +356,18 @@ static gboolean task_item_draw (
             (area.y + (area.height - gdk_pixbuf_get_height (desat)) / 2));
         g_object_unref (desat);
     }
-    if (!priv->mouse_over && attention) { /* urgent */
+    if (!item->mouse_over && attention) { /* urgent */
         GTimeVal current_time;
+        gdouble ms;
+        gdouble alpha;
+
         g_get_current_time (&current_time);
-        gdouble ms = (
-            current_time.tv_sec - priv->urgent_time.tv_sec) * 1000 +
-            (current_time.tv_usec - priv->urgent_time.tv_usec) / 1000;
-        gdouble alpha = .66 + (cos (3.15 * ms / 600) / 3);
+        ms = (
+            current_time.tv_sec - item->urgent_time.tv_sec) * 1000 +
+            (current_time.tv_usec - item->urgent_time.tv_usec) / 1000;
+        alpha = .66 + (cos (3.15 * ms / 600) / 3);
         cairo_paint_with_alpha (cr, alpha);
-    } else if (priv->mouse_over || active || !icons_greyscale) { /* focused */
+    } else if (item->mouse_over || active || !icons_greyscale) { /* focused */
         cairo_paint (cr);
     } else { /* not focused */
         cairo_paint_with_alpha (cr, .65);
@@ -357,14 +382,12 @@ static void on_size_allocate (
     TaskItem      *item)
 {
     g_return_if_fail (TASK_IS_ITEM (item));
-    TaskItemPrivate *priv;
     if (allocation->width != allocation->height + 6)
         gtk_widget_set_size_request (widget, allocation->height + 6, -1);
-    priv = item->priv;
-    priv->area.x = allocation->x;
-    priv->area.y = allocation->y;
-    priv->area.width = allocation->width;
-    priv->area.height = allocation->height;
+    item->area.x = allocation->x;
+    item->area.y = allocation->y;
+    item->area.width = allocation->width;
+    item->area.height = allocation->height;
     update_hints (item);
 }
 
@@ -375,7 +398,7 @@ static gboolean on_button_pressed (
 {
     WnckWindow *window;
     g_return_val_if_fail (TASK_IS_ITEM (item), FALSE);
-    window = item->priv->window;
+    window = item->window;
     g_return_val_if_fail (WNCK_IS_WINDOW (window), FALSE);
     if (event->button == 3) {
         GtkWidget *menu = wnck_action_menu_new (window);
@@ -395,7 +418,7 @@ static gboolean on_query_tooltip (
     GtkTooltip *tooltip,
     TaskItem *item)
 {
-    WnckWindow *window = item->priv->window;
+    WnckWindow *window = item->window;
     g_return_val_if_fail (WNCK_IS_WINDOW (window), FALSE);
     gtk_tooltip_set_text (tooltip, wnck_window_get_name(window));
     gtk_tooltip_set_icon (tooltip, wnck_window_get_icon (window));
@@ -408,7 +431,7 @@ static gboolean on_enter_notify (
     TaskItem *item)
 {
     g_return_val_if_fail (TASK_IS_ITEM (item), FALSE);
-    item->priv->mouse_over = TRUE;
+    item->mouse_over = TRUE;
     gtk_widget_queue_draw (widget);
     return FALSE;
 }
@@ -419,7 +442,7 @@ static gboolean on_leave_notify (
     TaskItem *item)
 {
     g_return_val_if_fail (TASK_IS_ITEM (item), FALSE);
-    item->priv->mouse_over = FALSE;
+    item->mouse_over = FALSE;
     gtk_widget_queue_draw (widget);
     return FALSE;
 }
@@ -427,10 +450,10 @@ static gboolean on_leave_notify (
 static gboolean on_blink (TaskItem *item) {
     g_return_val_if_fail (TASK_IS_ITEM (item), FALSE);
     gtk_widget_queue_draw (GTK_WIDGET (item));
-    if (wnck_window_or_transient_needs_attention (item->priv->window)) {
+    if (wnck_window_or_transient_needs_attention (item->window)) {
         return TRUE;
     } else {
-        item->priv->blink_timer = 0;
+        item->blink_timer = 0;
         return FALSE;
     }
 }
@@ -443,10 +466,9 @@ static void on_window_state_changed (
 {
     g_return_if_fail (WNCK_IS_WINDOW (window));
     g_return_if_fail (TASK_IS_ITEM (taskItem));
-    TaskItemPrivate *priv = taskItem->priv;
-    if (new_state & WNCK_WINDOW_STATE_URGENT && !priv->blink_timer) {
-        priv->blink_timer = g_timeout_add (30, (GSourceFunc)on_blink, taskItem);
-        g_get_current_time (&priv->urgent_time);
+    if (new_state & WNCK_WINDOW_STATE_URGENT && !taskItem->blink_timer) {
+        taskItem->blink_timer = g_timeout_add (30, (GSourceFunc) on_blink, taskItem);
+        g_get_current_time (&taskItem->urgent_time);
     }
     task_item_set_visibility (taskItem);
 }
@@ -460,11 +482,7 @@ static void on_window_workspace_changed (
 
 static void on_window_icon_changed (WnckWindow *window, TaskItem *item) {
     g_return_if_fail (TASK_IS_ITEM (item));
-    TaskItemPrivate *priv = item->priv;
-    if (GDK_IS_PIXBUF (priv->pixbuf)) {
-        g_object_unref (priv->pixbuf);
-        priv->pixbuf = NULL;
-    }
+    g_clear_object (&item->pixbuf);
     gtk_widget_queue_draw (GTK_WIDGET (item));
 }
 
@@ -474,7 +492,7 @@ on_window_type_changed (WnckWindow *window,
 {
     WnckWindowType type;
 
-    if (item->priv->window != window)
+    if (item->window != window)
         return;
 
     type = wnck_window_get_window_type (window);
@@ -517,10 +535,10 @@ on_window_geometry_changed (WnckWindow *window,
 
     window_monitor = get_window_monitor (window);
 
-    old_monitor = item->priv->monitor;
+    old_monitor = item->monitor;
     if (old_monitor != window_monitor)
       {
-        item->priv->monitor = window_monitor;
+        item->monitor = window_monitor;
 
         g_signal_emit (item, task_item_signals[TASK_ITEM_MONITOR_CHANGED], 0,
                        old_monitor);
@@ -532,11 +550,9 @@ static void on_screen_active_window_changed (
     WnckWindow    *old_window,
     TaskItem      *item)
 {
-    g_return_if_fail (TASK_IS_ITEM (item));
     WnckWindow *window;
-    TaskItemPrivate *priv = item->priv;
-    window = priv->window;
-    g_return_if_fail (WNCK_IS_WINDOW (window));
+    window = item->window;
+
     if ((WNCK_IS_WINDOW (old_window) && window == old_window) ||
         window == wnck_screen_get_active_window (screen))
     {
@@ -569,120 +585,147 @@ static void on_screen_window_closed (
     TaskItem     *item)
 {
     g_return_if_fail (TASK_IS_ITEM(item));
-    TaskItemPrivate *priv = item->priv;
-    g_return_if_fail (WNCK_IS_WINDOW (priv->window));
-    if (priv->window == window) {
+    g_return_if_fail (WNCK_IS_WINDOW (item->window));
+    if (item->window == window) {
         task_item_close (item);
     }
 }
 
-static void
-disconnect_screen (TaskItem *item)
-{
-    WnckScreen *screen;
-
-    screen = item->priv->screen;
-
-    g_signal_handlers_disconnect_by_func (screen, on_screen_window_closed,
-                                          item);
-    g_signal_handlers_disconnect_by_func (screen,
-                                          on_screen_active_window_changed,
-                                          item);
-    g_signal_handlers_disconnect_by_func (screen,
-                                          on_screen_active_workspace_changed,
-                                          item);
-    g_signal_handlers_disconnect_by_func (screen,
-                                          on_screen_active_viewport_changed,
-                                          item);
-}
-
-static void
-disconnect_window (TaskItem *item)
-{
-    WnckWindow *window;
-
-    window = item->priv->window;
-
-    g_signal_handlers_disconnect_by_func (window, on_window_workspace_changed,
-                                          item);
-    g_signal_handlers_disconnect_by_func (window, on_window_state_changed,
-                                          item);
-    g_signal_handlers_disconnect_by_func (window, on_window_icon_changed, item);
-    g_signal_handlers_disconnect_by_func (window, on_window_type_changed, item);
-    g_signal_handlers_disconnect_by_func (window, on_window_geometry_changed,
-                                          item);
-}
-
 static void
 task_item_close (TaskItem   *item)
 {
-    disconnect_window (item);
-    disconnect_screen (item);
-
     g_signal_emit (G_OBJECT (item),
                    task_item_signals[TASK_ITEM_CLOSED_SIGNAL], 0);
 }
 
 static gboolean activate_window (GtkWidget *widget) {
-    gint active;
-    TaskItemPrivate *priv;
-    g_return_val_if_fail (GTK_IS_WIDGET (widget), FALSE);
+    TaskItem *item;
+    GtkWidget *parent;
+    WnckWindow *window;
+    guint time;
+
     g_return_val_if_fail (TASK_IS_ITEM(widget), FALSE);
-    priv = TASK_ITEM (widget)->priv;
-    g_return_val_if_fail (WNCK_IS_WINDOW (priv->window), FALSE);
-    active = GPOINTER_TO_INT (
-        g_object_get_data (G_OBJECT (widget), "drag-true")
-    );
-    if (active) {
-        WnckWindow *window = priv->window;
-        if (WNCK_IS_WINDOW (window))
-            wnck_window_activate (window, time (NULL));
-    }
-    g_object_set_data (
-        G_OBJECT (widget), "drag-true", GINT_TO_POINTER (0));
+    item = TASK_ITEM (widget);
+    parent = gtk_widget_get_parent (widget);
+
+    time = GPOINTER_TO_INT (g_object_get_data (G_OBJECT (parent), "event-time"));
+
+    window = item->window;
+    if (WNCK_IS_WINDOW (window))
+        wnck_window_activate (window, time);
+
+    g_object_set_data (G_OBJECT (parent), "event-source", GINT_TO_POINTER (0));
+
     return FALSE;
 }
 
 /* Emitted when a drag leaves the destination */
 static void on_drag_leave (
-    GtkWidget *item,
+    GtkWidget *widget,
     GdkDragContext *context,
     guint time,
     gpointer user_data)
 {
-    g_object_set_data (G_OBJECT (item), "drag-true", GINT_TO_POINTER (0));
+  GtkWidget *parent;
+
+  parent = gtk_widget_get_parent (widget);
+  g_object_set_data (G_OBJECT (parent), "active-widget", NULL);
+
+  g_object_set_data (G_OBJECT (widget), "drag-true", GINT_TO_POINTER (0));
+}
+
+static gboolean
+on_drag_drop (GtkWidget      *widget,
+              GdkDragContext *context,
+              gint            x,
+              gint            y,
+              guint           time,
+              gpointer        user_data)
+{
+  return FALSE;
+}
+
+static gboolean
+find_drop_type_by_name (char *source_name)
+{
+  for (gulong i = 0; i < G_N_ELEMENTS (drop_types); i++)
+    {
+      gchar * target_name;
+
+      target_name = drop_types[i].target;
+
+      if (g_strcmp0 (source_name, target_name) == 0)
+        return TRUE;
+    }
+
+  return FALSE;
 }
 
 /* Emitted when a drag is over the destination */
-static gboolean on_drag_motion (
-    GtkWidget      *item,
-    GdkDragContext *context,
-    gint            x,
-    gint            y,
-    guint           time)
+static gboolean
+on_drag_motion (GtkWidget      *item,
+                GdkDragContext *context,
+                gint            x,
+                gint            y,
+                guint           time)
 {
-    GdkAtom         target_type = NULL;
+    GdkAtom target;
+    GList *targets;
+    GList *current;
+    GtkWidget *parent;
+    GtkWidget *active_widget;
+    guint event_source;
+
+    parent = gtk_widget_get_parent (item);
+    active_widget = g_object_get_data (G_OBJECT (parent), "active-widget");
+
+    if (active_widget == item)
+      return FALSE;
+
+    event_source = GPOINTER_TO_INT (g_object_get_data (G_OBJECT (parent), "event-source"));
+
+    if (event_source > 0)
+      {
+        g_source_remove (event_source);
+        g_object_set_data (G_OBJECT (parent), "event-source", GINT_TO_POINTER (0));
+      }
+
+    g_object_set_data (G_OBJECT (parent), "active-widget", item);
+
+    target = NULL;
+    targets = gdk_drag_context_list_targets (context);
+
+    if (targets == NULL)
+      return FALSE;
+
+    current = targets;
+
+    while (current != NULL)
+      {
+        char *name;
+        gboolean found;
 
-    if (gdk_drag_context_list_targets(context)) {
         /* Choose the best target type */
-        target_type = GDK_POINTER_TO_ATOM (
-            g_list_nth_data (
-                gdk_drag_context_list_targets(context),
-                    TARGET_WIDGET_DRAGGED
-            )
-        );
-        g_assert(target_type != NULL);
+        target = GDK_POINTER_TO_ATOM (current->data);
+        name = gdk_atom_name (target);
 
-        gtk_drag_get_data (
-            item,         // will receive 'drag-data-received' signal
-            context,        // represents the current state of the DnD
-            target_type,    // the target type we want
-            time            // time stamp
-        );
-    } else {
-        g_warning("Drag ended without target");
-    }
-    return FALSE;
+        found = find_drop_type_by_name (name);
+        g_free (name);
+
+        if (found)
+          break;
+
+        current = g_list_next (current);
+      }
+
+    g_assert (target != NULL);
+
+    gtk_drag_get_data (item,    // will receive 'drag-data-received' signal
+                       context, // represents the current state of the DnD
+                       target,  // the target type we want
+                       time);   // time stamp
+
+    return TRUE;
 }
 
 
@@ -693,63 +736,86 @@ static gboolean on_drag_motion (
  */
 static void on_drag_begin(GtkWidget *widget, GdkDragContext *context, gpointer user_data) {
     TaskItem *item = TASK_ITEM (widget);
-    TaskItemPrivate *priv = item->priv;
-    GdkRectangle area = priv->area;
+    GdkRectangle area = item->area;
     gint size = MIN (area.height, area.width);
-    GdkPixbuf *pixbuf = task_item_sized_pixbuf_for_window (item, priv->window, size);
+    GdkPixbuf *pixbuf = task_item_sized_pixbuf_for_window (item, item->window, size);
     gtk_drag_source_set_icon_pixbuf(widget, pixbuf);
     g_object_set_data (G_OBJECT (item), "drag-true", GINT_TO_POINTER (1));
 }
 
-static void on_drag_get_data(
-    GtkWidget *widget,
-    GdkDragContext *context,
-    GtkSelectionData *selection_data,
-    guint target_type,
-    guint time,
-    gpointer user_data)
+static void
+on_drag_get_data (GtkWidget        *widget,
+                  GdkDragContext   *context,
+                  GtkSelectionData *selection_data,
+                  guint             target_type,
+                  guint             time,
+                  gpointer          user_data)
 {
-    switch(target_type) {
-        case TARGET_WIDGET_DRAGGED:
-            g_assert(user_data != NULL && TASK_IS_ITEM(user_data));
-            gtk_selection_data_set(
-                selection_data,
-                gtk_selection_data_get_target (selection_data),
-                8,
-                (guchar*) &user_data,
-                sizeof(gpointer*)
-            );
-            break;
-        default:
-            g_assert_not_reached ();
-    }
+    TaskItem *item;
+
+    g_assert (user_data != NULL && TASK_IS_ITEM (user_data));
+
+    item = TASK_ITEM (user_data);
+
+    if (target_type != TARGET_TASK_ITEM_DRAGGED)
+      g_assert_not_reached ();
+
+    gtk_selection_data_set (selection_data,
+                            gtk_selection_data_get_target (selection_data),
+                            8,
+                            (guchar*) &item,
+                            sizeof (TaskItem*));
 }
 
-static void on_drag_end (
-    GtkWidget *widget,
-    GdkDragContext *drag_context,
-    gpointer user_data)
+static void
+on_drag_end (GtkWidget      *drag_source,
+             GdkDragContext *drag_context,
+             gpointer        user_data)
 {
-    g_object_set_data (G_OBJECT (widget), "drag-true", GINT_TO_POINTER (0));
+  GtkWidget *parent;
+  guint event_source;
+
+  parent = gtk_widget_get_parent (drag_source);
+
+  g_object_set_data (G_OBJECT (parent), "active-widget", NULL);
+  event_source = GPOINTER_TO_INT (g_object_get_data (G_OBJECT (parent), "event-source"));
+
+  if (event_source != 0)
+  {
+    g_source_remove (event_source);
+    g_object_set_data (G_OBJECT (parent), "event-source", GINT_TO_POINTER (0));
+  }
+
+  g_object_set_data (G_OBJECT (drag_source), "drag-true", GINT_TO_POINTER (0));
 }
 
-static gint grid_get_pos (GtkWidget *grid, GtkWidget *item) {
-    GtkContainer *container = GTK_CONTAINER (grid);
-    GList *items = gtk_container_get_children (container);
+static gint
+grid_get_pos (GtkWidget *grid,
+              GtkWidget *item)
+{
+  GtkContainer *container;
+  GList *items;
+
+  container = GTK_CONTAINER (grid);
+  items = gtk_container_get_children (container);
+
+    while (items)
+      {
+        if (items->data == item)
+          {
+            gint pos;
+            gtk_container_child_get (container, item, "position", &pos, NULL);
+            return pos;
+          }
 
-    while (items) {
-        if (items->data == item) {
-	        gint pos;
-	        gtk_container_child_get (container, item, "position", &pos, NULL);
-	        return pos;
-	    }
         items = items->next;
-    }
+      }
+
     return -1;
 }
 
 static void on_drag_received_data (
-    GtkWidget *widget, //target of the d&d action
+    GtkWidget *drag_target, //target of the d&d action
     GdkDragContext *context,
     gint x,
     gint y,
@@ -758,40 +824,71 @@ static void on_drag_received_data (
     guint time,
     TaskItem *item)
 {
-    if((selection_data != NULL) && (gtk_selection_data_get_length(selection_data) >= 0)) {
-        gint active;
-        switch (target_type) {
-            case TARGET_WIDGET_DRAGGED: {
-                GtkWidget *taskList = wp_applet_get_tasks(item->priv->windowPickerApplet);
-                gpointer *data = (gpointer *) gtk_selection_data_get_data(selection_data);
-                g_assert(GTK_IS_WIDGET(*data));
-
-                GtkWidget *taskItem = GTK_WIDGET(*data);
-                g_assert(TASK_IS_ITEM(taskItem));
-                if(taskItem == widget) break; //source and target are identical
-                gint target_position = grid_get_pos(wp_applet_get_tasks(item->priv->windowPickerApplet), widget);
-                g_object_ref(taskItem);
-                gtk_box_reorder_child(GTK_BOX(taskList), taskItem, target_position);
-                g_object_unref(taskItem);
-                break;
-            }
-            default:
-                active = GPOINTER_TO_INT (g_object_get_data (G_OBJECT (widget), "drag-true"));
-                if (!active) {
-                    g_object_set_data (
-                        G_OBJECT (widget), "drag-true", GINT_TO_POINTER (1)
-                    );
-                    g_timeout_add (1000, (GSourceFunc)activate_window, widget);
-                }
-        }
+  if (selection_data == NULL || gtk_selection_data_get_length (selection_data) < 0)
+    {
+      gdk_drag_status (context, 0, time);
+      return;
     }
+
+  if (target_type == TARGET_TASK_ITEM_DRAGGED)
+    {
+      GtkWidget *drag_source;
+      GtkWidget *taskList;
+      GtkWidget *parent;
+      gint target_position;
+      gpointer *data;
+
+      parent = gtk_widget_get_parent (drag_target);
+      taskList = wp_applet_get_tasks (item->windowPickerApplet);
+      data = (gpointer *) gtk_selection_data_get_data (selection_data);
+
+      drag_source = GTK_WIDGET (*data);
+      g_assert (TASK_IS_ITEM (drag_source));
+
+      if (drag_source == drag_target)
+        {
+          gdk_drag_status (context, 0, time);
+          return; //source and target are identical
+        }
+
+      target_position = grid_get_pos (taskList, drag_target);
+
+      g_object_ref (drag_source);
+      gtk_box_reorder_child (GTK_BOX (taskList), drag_source, target_position);
+      g_object_unref (drag_source);
+
+      // we swapped the source and the target, so the drag_target is no longer part of the drag action
+      g_object_set_data (G_OBJECT (drag_target), "drag-true",
+                         GINT_TO_POINTER (0));
+
+      // we swapped the source and the target, so the drag_source is the new active widget
+      g_object_set_data (G_OBJECT (parent), "active-widget", drag_source);
+    }
+  else
+    {
+      GtkWidget *parent;
+      guint event_source;
+
+      parent = gtk_widget_get_parent (drag_target);
+      event_source = g_timeout_add (1000,
+                                    (GSourceFunc) activate_window,
+                                    drag_target);
+
+      g_object_set_data (G_OBJECT (parent), "event-source",
+                         GINT_TO_POINTER (event_source));
+
+      g_object_set_data (G_OBJECT (parent), "event-time",
+                         GINT_TO_POINTER (time));
+    }
+
+  gdk_drag_status (context, GDK_ACTION_COPY, time);
 }
 
 /* Returning true here, causes the failed-animation not to be shown. Without this the icon of the dnd operation
  * will jump back to where the dnd operation started.
  **/
 static gboolean
-on_drag_failed (GtkWidget      *widget,
+on_drag_failed (GtkWidget      *drag_source,
                 GdkDragContext *context,
                 GtkDragResult   result,
                 TaskItem       *taskItem)
@@ -800,14 +897,12 @@ on_drag_failed (GtkWidget      *widget,
 }
 
 static void task_item_setup_atk (TaskItem *item) {
-    TaskItemPrivate *priv;
     GtkWidget *widget;
     AtkObject *atk;
     WnckWindow *window;
     g_return_if_fail (TASK_IS_ITEM (item));
     widget = GTK_WIDGET (item);
-    priv = item->priv;
-    window = priv->window;
+    window = item->window;
     g_return_if_fail (WNCK_IS_WINDOW (window));
     atk = gtk_widget_get_accessible (widget);
     atk_object_set_name (atk, _("Window Task Button"));
@@ -815,42 +910,33 @@ static void task_item_setup_atk (TaskItem *item) {
     atk_object_set_role (atk, ATK_ROLE_PUSH_BUTTON);
 }
 
-static void
-task_item_dispose (GObject *object)
-{
-    TaskItem *task_item = TASK_ITEM (object);
-
-    disconnect_window (task_item);
-    disconnect_screen (task_item);
-
-    G_OBJECT_CLASS (task_item_parent_class)->dispose (object);
-}
-
 static void task_item_finalize (GObject *object) {
-    TaskItemPrivate *priv = TASK_ITEM (object)->priv;
-    if (priv->blink_timer) {
-        g_source_remove (priv->blink_timer);
+    TaskItem *item;
+
+    item = TASK_ITEM (object);
+
+    if (item->blink_timer) {
+        g_source_remove (item->blink_timer);
     }
 
-    if (GDK_IS_PIXBUF (priv->pixbuf)) {
-        g_object_unref (priv->pixbuf);
-    }
+    g_clear_object (&item->pixbuf);
+
     G_OBJECT_CLASS (task_item_parent_class)->finalize (object);
 }
 
 static void task_item_class_init (TaskItemClass *klass) {
     GObjectClass *obj_class      = G_OBJECT_CLASS (klass);
     GtkWidgetClass *widget_class = GTK_WIDGET_CLASS (klass);
-    obj_class->dispose = task_item_dispose;
+
     obj_class->finalize = task_item_finalize;
+
     widget_class->get_preferred_width = task_item_get_preferred_width;
     widget_class->get_preferred_height = task_item_get_preferred_height;
-    g_type_class_add_private (obj_class, sizeof (TaskItemPrivate));
     task_item_signals [TASK_ITEM_CLOSED_SIGNAL] =
     g_signal_new ("task-item-closed",
         G_TYPE_FROM_CLASS (klass),
         G_SIGNAL_RUN_FIRST | G_SIGNAL_ACTION,
-        G_STRUCT_OFFSET (TaskItemClass, itemclosed),
+        0,
         NULL, NULL,
         g_cclosure_marshal_VOID__VOID, G_TYPE_NONE, 0);
 
@@ -860,16 +946,17 @@ static void task_item_class_init (TaskItemClass *klass) {
 }
 
 static void task_item_init (TaskItem *item) {
-    TaskItemPrivate *priv = item->priv = TASK_ITEM_GET_PRIVATE (item);
-    priv->blink_timer = 0;
+    item->blink_timer = 0;
 }
 
 GtkWidget *task_item_new (WpApplet* windowPickerApplet, WnckWindow *window) {
-    g_return_val_if_fail (WNCK_IS_WINDOW (window), NULL);
     TaskItem *taskItem;
-    TaskItemPrivate *priv;
     WnckScreen *screen;
-    GtkWidget *item = g_object_new (
+    GtkWidget *item;
+
+    g_return_val_if_fail (WNCK_IS_WINDOW (window), NULL);
+
+    item = g_object_new (
         TASK_TYPE_ITEM,
         "has-tooltip", TRUE,
         "visible-window", FALSE,
@@ -880,12 +967,11 @@ GtkWidget *task_item_new (WpApplet* windowPickerApplet, WnckWindow *window) {
     gtk_widget_add_events (item, GDK_ALL_EVENTS_MASK);
     gtk_container_set_border_width (GTK_CONTAINER (item), 0);
     taskItem = TASK_ITEM (item);
-    priv = taskItem->priv;
-    priv->window = window;
+    taskItem->window = window;
     screen = wnck_window_get_screen (window);
-    priv->screen = screen;
-    priv->windowPickerApplet = windowPickerApplet;
-    priv->monitor = get_window_monitor (window);
+    taskItem->screen = screen;
+    taskItem->windowPickerApplet = windowPickerApplet;
+    taskItem->monitor = get_window_monitor (window);
 
     /** Drag and Drop code
      * This item can be both the target and the source of a drag and drop
@@ -896,7 +982,7 @@ GtkWidget *task_item_new (WpApplet* windowPickerApplet, WnckWindow *window) {
     //target (destination)
     gtk_drag_dest_set (
         item,
-        GTK_DEST_DEFAULT_HIGHLIGHT,
+        GTK_DEST_DEFAULT_MOTION,
         drop_types, n_drop_types,
         GDK_ACTION_COPY
     );
@@ -917,6 +1003,8 @@ GtkWidget *task_item_new (WpApplet* windowPickerApplet, WnckWindow *window) {
         G_CALLBACK (on_drag_motion), item);
     g_signal_connect (item, "drag-leave",
         G_CALLBACK (on_drag_leave), item);
+    g_signal_connect (item, "drag-drop",
+        G_CALLBACK (on_drag_drop), item);
     g_signal_connect (item, "drag_data_received",
         G_CALLBACK(on_drag_received_data), item);
     /* a 'drag-drop' signal is not needed, instead we rely on the drag-failed signal to end a drag operation. */
@@ -931,24 +1019,25 @@ GtkWidget *task_item_new (WpApplet* windowPickerApplet, WnckWindow *window) {
         G_CALLBACK (on_drag_get_data), item);
 
     /* Other signals */
-    g_signal_connect (screen, "viewports-changed",
-        G_CALLBACK (on_screen_active_viewport_changed), item);
-    g_signal_connect (screen, "active-window-changed",
-        G_CALLBACK (on_screen_active_window_changed), item);
-    g_signal_connect (screen, "active-workspace-changed",
-        G_CALLBACK (on_screen_active_workspace_changed), item);
-    g_signal_connect (screen, "window-closed",
-        G_CALLBACK (on_screen_window_closed), item);
-    g_signal_connect (window, "workspace-changed",
-        G_CALLBACK (on_window_workspace_changed), item);
-    g_signal_connect (window, "state-changed",
-        G_CALLBACK (on_window_state_changed), item);
-    g_signal_connect (window, "icon-changed",
-        G_CALLBACK (on_window_icon_changed), item);
-    g_signal_connect (window, "type-changed",
-        G_CALLBACK (on_window_type_changed), item);
-    g_signal_connect (window, "geometry-changed",
-        G_CALLBACK (on_window_geometry_changed), item);
+    g_signal_connect_object (screen, "viewports-changed",
+        G_CALLBACK (on_screen_active_viewport_changed), item, 0);
+    g_signal_connect_object (screen, "active-window-changed",
+        G_CALLBACK (on_screen_active_window_changed), item, 0);
+    g_signal_connect_object (screen, "active-workspace-changed",
+        G_CALLBACK (on_screen_active_workspace_changed), item, 0);
+    g_signal_connect_object (screen, "window-closed",
+        G_CALLBACK (on_screen_window_closed), item, 0);
+    g_signal_connect_object (window, "workspace-changed",
+        G_CALLBACK (on_window_workspace_changed), item, 0);
+    g_signal_connect_object (window, "state-changed",
+        G_CALLBACK (on_window_state_changed), item, 0);
+    g_signal_connect_object (window, "icon-changed",
+        G_CALLBACK (on_window_icon_changed), item, 0);
+    g_signal_connect_object (window, "type-changed",
+        G_CALLBACK (on_window_type_changed), item, 0);
+    g_signal_connect_object (window, "geometry-changed",
+        G_CALLBACK (on_window_geometry_changed), item, 0);
+
     g_signal_connect(item, "draw",
         G_CALLBACK(task_item_draw), windowPickerApplet);
     g_signal_connect (item, "button-release-event",
@@ -971,5 +1060,5 @@ GtkWidget *task_item_new (WpApplet* windowPickerApplet, WnckWindow *window) {
 GdkMonitor *
 task_item_get_monitor (TaskItem *item)
 {
-    return item->priv->monitor;
+    return item->monitor;
 }
